import os
import sys

current_directory: str = os.path.dirname(os.path.abspath(__file__))
if current_directory.endswith("tests") and current_directory not in sys.path:
    sys.path.append(current_directory)



from __future__ import annotations

from copy import deepcopy
from typing import List, Optional

from clp_ffi_py.ir.native import Query
from clp_ffi_py.wildcard_query import WildcardQuery


class QueryBuilderException(Exception):
    """
    Exception raised when building a :class:`~clp_ffi_py.ir.native.Query` fails.
    """

    pass


class QueryBuilder:
    """
    This class serves as an interface for conveniently constructing Query
    objects utilized in CLP IR streaming search. It provides methods for
    configuring and resetting search parameters.

    For more details about the search query CLP IR stream supports, see
    :class:`~clp_ffi_py.ir.native.Query` and
    :class:`~clp_ffi_py.wildcard_query.WildcardQuery`.
    """

    def __init__(self) -> None:
        self._search_time_lower_bound: int = Query.default_search_time_lower_bound()
        self._search_time_upper_bound: int = Query.default_search_time_upper_bound()
        self._search_time_termination_margin: int = Query.default_search_time_termination_margin()
        self._wildcard_queries: List[WildcardQuery] = []

    @property
    def search_time_lower_bound(self) -> int:
        return self._search_time_lower_bound

    @property
    def search_time_upper_bound(self) -> int:
        return self._search_time_upper_bound

    @property
    def search_time_termination_margin(self) -> int:
        return self._search_time_termination_margin

    @property
    def wildcard_queries(self) -> List[WildcardQuery]:
        """
        :return: A deep copy of the underlying wildcard query list.
        """
        return deepcopy(self._wildcard_queries)

    def set_search_time_lower_bound(self, ts: int) -> QueryBuilder:
        """
        :param ts: Start of the search time range (inclusive) as a UNIX epoch
            timestamp in milliseconds.
        :return: self.
        """
        self._search_time_lower_bound = ts
        return self

    def set_search_time_upper_bound(self, ts: int) -> QueryBuilder:
        """
        :param ts: End of the search time range (inclusive) as a UNIX epoch
            timestamp in milliseconds.
        :return: self.
        """
        self._search_time_upper_bound = ts
        return self

    def set_search_time_termination_margin(self, ts: int) -> QueryBuilder:
        """
        :param ts: The search time termination margin as a UNIX epoch timestamp
            in milliseconds.
        :return: self.
        """
        self._search_time_termination_margin = ts
        return self

    def add_wildcard_query(self, wildcard_query: str, case_sensitive: bool = False) -> QueryBuilder:
        """
        Constructs and adds a :class:`~clp_ffi_py.wildcard_query.WildcardQuery`
        to the wildcard query list.

        :param wildcard_query: The wildcard query string to add.
        :param case_sensitive: Whether to perform case-sensitive matching.
        :return: self.
        """
        self._wildcard_queries.append(WildcardQuery(wildcard_query, case_sensitive))
        return self

    def add_wildcard_queries(self, wildcard_queries: List[WildcardQuery]) -> QueryBuilder:
        """
        Adds a list of wildcard queries to the wildcard query list.

        :param wildcard_queries: The list of wildcard queries to add.
        :return: self.
        """
        self._wildcard_queries.extend(wildcard_queries)
        return self

    def reset_search_time_lower_bound(self) -> QueryBuilder:
        """
        Resets the search time lower bound to the default value.

        :return: self.
        """
        self._search_time_lower_bound = Query.default_search_time_lower_bound()
        return self

    def reset_search_time_upper_bound(self) -> QueryBuilder:
        """
        Resets the search time upper bound to the default value.

        :return: self.
        """
        self._search_time_upper_bound = Query.default_search_time_upper_bound()
        return self

    def reset_search_time_termination_margin(self) -> QueryBuilder:
        """
        Resets the search time termination margin to the default value.

        :return: self.
        """
        self._search_time_termination_margin = Query.default_search_time_termination_margin()
        return self

    def reset_wildcard_queries(self) -> QueryBuilder:
        """
        Clears the wildcard query list.

        :return: self.
        """
        self._wildcard_queries.clear()
        return self

    def reset(self) -> QueryBuilder:
        """
        Resets all settings to their defaults.

        :return: self.
        """
        return (
            self.reset_wildcard_queries()
            .reset_search_time_termination_margin()
            .reset_search_time_upper_bound()
            .reset_search_time_lower_bound()
        )

    def build(self) -> Query:
        """
        :raises QueryBuilderException: If the search time range lower bound
            exceeds the search time range upper bound.
        :return: A :class:`~clp_ffi_py.ir.native.Query` object initialized
            with the parameters set by the builder.
        """
        if self._search_time_lower_bound > self._search_time_upper_bound:
            raise QueryBuilderException(
                "The search time lower bound exceeds the search time upper bound."
            )
        wildcard_queries: Optional[List[WildcardQuery]] = None
        if 0 != len(self._wildcard_queries):
            wildcard_queries = self._wildcard_queries
        return Query(
            search_time_lower_bound=self._search_time_lower_bound,
            search_time_upper_bound=self._search_time_upper_bound,
            search_time_termination_margin=self._search_time_termination_margin,
            wildcard_queries=wildcard_queries,
        )



import io
import random
from pathlib import Path
from typing import Optional

from smart_open import open  # type: ignore
from test_ir.test_utils import TestCLPBase

from clp_ffi_py.ir import DecoderBuffer


class TestCaseDecoderBuffer(TestCLPBase):
    """
    Class for testing clp_ffi_py.ir.DecoderBuffer.
    """

    input_src_dir: str = "test_data"

    def test_buffer_protocol(self) -> None:
        """
        Tests whether the buffer protocol is disabled by default.
        """
        byte_array: bytearray = bytearray(b"Hello, world!")
        byte_stream: io.BytesIO = io.BytesIO(byte_array)
        decoder_buffer: DecoderBuffer = DecoderBuffer(byte_stream)
        exception_captured: bool = False
        try:
            byte_stream.readinto(decoder_buffer)  # type: ignore
        except TypeError:
            exception_captured = True
        self.assertTrue(
            exception_captured, "The buffer protocol should not be enabled in Python layer."
        )

    def test_streaming_small_buffer(self) -> None:
        """
        Tests DecoderBuffer's functionality using the small buffer capacity.
        """
        buffer_capacity: int = 1024
        self.__launch_test(buffer_capacity)

    def test_streaming_default_buffer(self) -> None:
        """
        Tests DecoderBuffer's functionality using the default buffer capacity.
        """
        self.__launch_test(None)

    def test_streaming_large_buffer(self) -> None:
        """
        Tests DecoderBuffer's functionality using the large buffer capacity.
        """
        buffer_capacity: int = 16384
        self.__launch_test(buffer_capacity)

    def __launch_test(self, buffer_capacity: Optional[int]) -> None:
        """
        Tests the DecoderBuffer by streaming the files inside `test_src_dir`.

        :param self
        :param buffer_capacity: The buffer capacity used to initialize the
            decoder buffer.
        """
        current_dir: Path = Path(__file__).resolve().parent
        test_src_dir: Path = current_dir / TestCaseDecoderBuffer.input_src_dir
        for file_path in test_src_dir.rglob("*"):
            if not file_path.is_file():
                continue
            streaming_result: bytearray
            decoder_buffer: DecoderBuffer
            random_seed: int
            # Run against 10 different seeds:
            for _ in range(10):
                random_seed = random.randint(1, 3190)
                with open(str(file_path), "rb") as istream:
                    try:
                        if None is buffer_capacity:
                            decoder_buffer = DecoderBuffer(istream)
                        else:
                            decoder_buffer = DecoderBuffer(
                                initial_buffer_capacity=buffer_capacity, input_stream=istream
                            )
                        streaming_result = decoder_buffer._test_streaming(random_seed)
                    except Exception as e:
                        self.assertFalse(
                            True, f"Error on file {file_path} using seed {random_seed}: {e}"
                        )
                self.__assert_streaming_result(file_path, streaming_result, random_seed)

    def __assert_streaming_result(
        self, file_path: Path, streaming_result: bytearray, random_seed: int
    ) -> None:
        """
        Validates the streaming result read by the decoder buffer.

        :param file_path: Input stream file Path.
        :param streaming_result: Result of DecoderBuffer `_test_streaming` method.
        """
        with open(str(file_path), "rb") as istream:
            ref_result: bytearray = bytearray(istream.read())
            self.assertEqual(
                ref_result,
                streaming_result,
                f"Streaming result is different from the src: {file_path}. Random seed:"
                f" {random_seed}.",
            )



# -- Project information -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "clp-ffi-py"
# NOTE: We don't include a period after "Inc" since the theme adds one already.
copyright = "2023-2024 YScope Inc"

# -- General configuration -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_design",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx_copybutton",
]

# -- MyST extensions -----------------------------------------------------------
# https://myst-parser.readthedocs.io/en/stable/syntax/optional.html
myst_enable_extensions = [
    "colon_fence",
]

# -- Sphinx autodoc options ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

autoclass_content = "class"
autodoc_class_signature = "separated"
autodoc_typehints = "description"

# -- HTML output options -------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_favicon = "https://docs.yscope.com/_static/favicon.ico"
html_title = "clp-ffi-py"
html_show_copyright = True

html_static_path = ["../src/_static"]

html_theme = "pydata_sphinx_theme"

# -- Theme options -------------------------------------------------------------
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/layout.html

html_theme_options = {
    "footer_start": ["copyright"],
    "footer_center": [],
    "footer_end": ["theme-version"],
    "primary_sidebar_end": [],
    "secondary_sidebar_items": ["page-toc", "edit-this-page"],
    "show_prev_next": False,
    "use_edit_page_button": True,
}

# -- Theme source buttons ------------------------------------------------------
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/source-buttons.html

html_context = {
    "github_user": "y-scope",
    "github_repo": "clp-ffi-py",
    "github_version": "main",
    "doc_path": "docs/src",
}

# -- Theme custom CSS and JS ---------------------------------------------------
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/static_assets.html

def setup(app):
    app.add_css_file("custom.css")



from __future__ import annotations

from pathlib import Path
from sys import stderr
from types import TracebackType
from typing import Generator, IO, Iterator, Optional, Type, Union

from zstandard import ZstdDecompressionReader, ZstdDecompressor

from clp_ffi_py.ir.native import Decoder, DecoderBuffer, LogEvent, Metadata, Query


class ClpIrStreamReader(Iterator[LogEvent]):
    """
    This class represents a stream reader used to read/decode encoded log events
    from a CLP IR stream. It also provides method(s) to instantiate a log event
    generator with a customized search query.

    :param istream: Input stream that contains encoded CLP IR.
    :param decoder_buffer_size: Initial size of the decoder buffer.
    :param enable_compression: A flag indicating whether the istream is
        compressed using `zstd`.
    :param allow_incomplete_stream: If set to `True`, an incomplete CLP IR
        stream is not treated as an error. Instead, encountering such a stream
        is seen as reaching its end without raising any exceptions.
    """

    DEFAULT_DECODER_BUFFER_SIZE: int = 65536

    def __init__(
        self,
        istream: IO[bytes],
        decoder_buffer_size: int = DEFAULT_DECODER_BUFFER_SIZE,
        enable_compression: bool = True,
        allow_incomplete_stream: bool = False,
    ):
        self.__istream: Union[IO[bytes], ZstdDecompressionReader]
        if enable_compression:
            dctx: ZstdDecompressor = ZstdDecompressor()
            self.__istream = dctx.stream_reader(istream, read_across_frames=True)
        else:
            self.__istream = istream
        self._decoder_buffer: DecoderBuffer = DecoderBuffer(self.__istream, decoder_buffer_size)
        self._metadata: Optional[Metadata] = None
        self._allow_incomplete_stream: bool = allow_incomplete_stream

    def read_next_log_event(self) -> Optional[LogEvent]:
        """
        Reads and decodes the next encoded log event from the IR stream.

        :return:
            - Next unread log event represented as an instance of LogEvent.
            - None if the end of IR stream is reached.
        :raise Exception:
            If :meth:`~clp_ffi_py.ir.native.Decoder.decode_next_log_event`
            fails.
        """
        return Decoder.decode_next_log_event(
            self._decoder_buffer, allow_incomplete_stream=self._allow_incomplete_stream
        )

    def read_preamble(self) -> None:
        """
        Try to decode the preamble and set `metadata`. If `metadata` has been
        set already, it will instantly return. It is separated from `__init__`
        so that the input stream does not need to be readable on a reader's
        construction, but until the user starts to iterate logs.

        :raise Exception:
            If :meth:`~clp_ffi_py.ir.native.Decoder.decode_preamble` fails.
        """
        if self.has_metadata():
            return
        self._metadata = Decoder.decode_preamble(self._decoder_buffer)

    def get_metadata(self) -> Metadata:
        if None is self._metadata:
            raise RuntimeError("The metadata has not been successfully decoded yet.")
        return self._metadata

    def has_metadata(self) -> bool:
        return None is not self._metadata

    def search(self, query: Query) -> Generator[LogEvent, None, None]:
        """
        Searches and yields log events that match a specific search query.

        :param query: The input query object used to match log events. Check the
            document of :class:`~clp_ffi_py.ir.Query` for more details.
        :yield: The next unread encoded log event that matches the given search
            query from the IR stream.
        """
        if False is self.has_metadata():
            self.read_preamble()
        while True:
            log_event: Optional[LogEvent] = Decoder.decode_next_log_event(
                self._decoder_buffer,
                query=query,
                allow_incomplete_stream=self._allow_incomplete_stream,
            )
            if None is log_event:
                break
            yield log_event

    def close(self) -> None:
        self.__istream.close()

    def __iter__(self) -> ClpIrStreamReader:
        if False is self.has_metadata():
            self.read_preamble()
        return self

    def __enter__(self) -> ClpIrStreamReader:
        if False is self.has_metadata():
            self.read_preamble()
        return self

    def __next__(self) -> LogEvent:
        next_log_event: Optional[LogEvent] = self.read_next_log_event()
        if None is next_log_event:
            raise StopIteration
        return next_log_event

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()


class ClpIrFileReader(ClpIrStreamReader):
    """
    Wrapper class of `ClpIrStreamReader` that calls `open` for convenience.
    """

    def __init__(
        self,
        fpath: Path,
        decoder_buffer_size: int = ClpIrStreamReader.DEFAULT_DECODER_BUFFER_SIZE,
        enable_compression: bool = True,
        allow_incomplete_stream: bool = False,
    ):
        self._path: Path = fpath
        super().__init__(
            open(fpath, "rb"),
            decoder_buffer_size=decoder_buffer_size,
            enable_compression=enable_compression,
            allow_incomplete_stream=allow_incomplete_stream,
        )

    def dump(self, ostream: IO[str] = stderr) -> None:
        for log_event in self:
            ostream.write(str(log_event))



import random
from pathlib import Path
from typing import List, Optional, Tuple

from smart_open import open  # type: ignore
from test_ir.test_utils import get_current_timestamp, LogGenerator, TestCLPBase

from clp_ffi_py.ir import (
    Decoder,
    DecoderBuffer,
    FourByteEncoder,
    LogEvent,
    Metadata,
    Query,
)
from clp_ffi_py.wildcard_query import WildcardQuery

LOG_DIR: Path = Path("unittest-logs")


class TestCaseDecoderBase(TestCLPBase):
    """
    Class for testing clp_ffi_py.ir.Decoder.
    """

    encoded_log_path_prefix: str
    encoded_log_path_postfix: str
    num_test_iterations: int
    enable_compression: bool
    has_query: bool

    # override
    @classmethod
    def setUpClass(cls) -> None:
        if not LOG_DIR.exists():
            LOG_DIR.mkdir(parents=True, exist_ok=True)
        assert LOG_DIR.is_dir()

    # override
    def setUp(self) -> None:
        self.encoded_log_path_prefix: str = f"{self.id()}"
        self.encoded_log_path_postfix: str = "clp.zst" if self.enable_compression else "clp"
        for i in range(self.num_test_iterations):
            log_path = self._get_log_path(i)
            if log_path.exists():
                log_path.unlink()

    def _get_log_path(self, iter: int) -> Path:
        log_path: Path = LOG_DIR / Path(
            f"{self.encoded_log_path_prefix}.{iter}.{self.encoded_log_path_postfix}"
        )
        return log_path

    def _encode_log_stream(
        self, log_path: Path, metadata: Metadata, log_events: List[LogEvent]
    ) -> None:
        """
        Encodes the log stream into the given path.

        :param log_path: Path on the local file system to write the stream.
        :param metadata: Metadata of the log stream.
        :param log_events: A list of log events to encode.
        """
        with open(str(log_path), "wb") as ostream:
            ref_timestamp: int = metadata.get_ref_timestamp()
            ostream.write(
                FourByteEncoder.encode_preamble(
                    ref_timestamp, metadata.get_timestamp_format(), metadata.get_timezone_id()
                )
            )
            for log_event in log_events:
                curr_ts: int = log_event.get_timestamp()
                delta: int = curr_ts - ref_timestamp
                ref_timestamp = curr_ts
                log_message: str = log_event.get_log_message()
                ostream.write(
                    FourByteEncoder.encode_message_and_timestamp_delta(delta, log_message.encode())
                )
            ostream.write(FourByteEncoder.encode_end_of_ir())

    def _encode_random_log_stream(
        self, log_path: Path, num_log_events_to_generate: int, seed: int
    ) -> Tuple[Metadata, List[LogEvent]]:
        """
        Writes a randomly generated log stream into the local path `log_path`.

        :param log_path: Path on the local file system to write the stream.
        :param num_log_events_to_generate: Number of log events to generate.
        :param seed: Random seed used to generate the log stream.
        :return: A tuple containing the generated log events and the metadata.
        """
        metadata: Metadata
        log_events: List[LogEvent]
        metadata, log_events = LogGenerator.generate_random_logs(num_log_events_to_generate)
        try:
            self._encode_log_stream(log_path, metadata, log_events)
        except Exception as e:
            self.assertTrue(
                False, f"Failed to encode random log stream generated using seed {seed}: {e}"
            )
        return metadata, log_events

    def _generate_random_query(
        self, ref_log_events: List[LogEvent]
    ) -> Tuple[Query, List[LogEvent]]:
        """
        Generates a random query and return all the log events in the given
        `log_events` that matches this random query.

        The derived class might overwrite this method to generate a random query
        using customized algorithm. By default, this function returns an empty
        query and `ref_log_events`.
        :param log_events: reference log events.
        :return: A tuple that contains the randomly generated query, and a list
        of log events filtered from `ref_log_events` by the query.
        """
        return Query(), ref_log_events

    def _decode_log_stream(
        self, log_path: Path, query: Optional[Query]
    ) -> Tuple[Metadata, List[LogEvent]]:
        """
        Decodes the log stream specified by `log_path`, using decoding methods
        provided in clp_ffi_py.ir.Decoder.

        :param log_path: The path to the log stream.
        :param query: Optional search query.
        :return: A tuple that contains the decoded metadata and log events
            returned from decoding methods.
        """
        with open(str(log_path), "rb") as istream:
            decoder_buffer: DecoderBuffer = DecoderBuffer(istream)
            metadata: Metadata = Decoder.decode_preamble(decoder_buffer)
            log_events: List[LogEvent] = []
            while True:
                log_event: Optional[LogEvent] = Decoder.decode_next_log_event(decoder_buffer, query)
                if None is log_event:
                    break
                log_events.append(log_event)
        return metadata, log_events

    def _validate_decoded_logs(
        self,
        ref_metadata: Metadata,
        ref_log_events: List[LogEvent],
        decoded_metadata: Metadata,
        decoded_log_events: List[LogEvent],
        log_path: Path,
        seed: int,
    ) -> None:
        """
        Validates decoded logs from the IR stream specified by `log_path`.

        :param ref_metadata: Reference metadata.
        :param ref_log_events: A list of reference log events sequence (order
            sensitive).
        :param decoded_metadata: Metadata decoded from the IR stream.
        :param decoded_log_events: A list of log events decoded from the IR
            stream in sequence.
        :param log_path: Local path of the IR stream.
        :param seed: Random seed used to generate the log events sequence.
        """
        test_info: str = f"Seed: {seed}, Log Path: {log_path}"
        self._check_metadata(
            decoded_metadata,
            ref_metadata.get_ref_timestamp(),
            ref_metadata.get_timestamp_format(),
            ref_metadata.get_timezone_id(),
            test_info,
        )

        ref_num_log_events: int = len(ref_log_events)
        decoded_num_log_events: int = len(decoded_log_events)
        self.assertEqual(
            ref_num_log_events,
            decoded_num_log_events,
            "Number of log events decoded does not match.\n" + test_info,
        )
        for ref_log_event, decoded_log_event in zip(ref_log_events, decoded_log_events):
            self._check_log_event(
                decoded_log_event,
                ref_log_event.get_log_message(),
                ref_log_event.get_timestamp(),
                ref_log_event.get_index(),
                test_info,
            )

    def test_decoder_with_random_logs(self) -> None:
        """
        Tests encoding/decoding methods.

        Check the TestCase class doc string for more details.
        """
        for i in range(self.num_test_iterations):
            seed: int = get_current_timestamp()
            random.seed(seed)
            num_log_events: int = 100 * (i + 1)
            log_path: Path = self._get_log_path(i)

            ref_metadata: Metadata
            ref_log_events: List[LogEvent]
            ref_metadata, ref_log_events = self._encode_random_log_stream(
                log_path, num_log_events, seed
            )

            query: Optional[Query] = None
            if self.has_query:
                query, ref_log_events = self._generate_random_query(ref_log_events)

            metadata: Metadata
            log_events: List[LogEvent]
            try:
                metadata, log_events = self._decode_log_stream(log_path, query)
            except Exception as e:
                self.assertTrue(
                    False, f"Failed to decode random log stream generated using seed {seed}: {e}"
                )

            self._validate_decoded_logs(
                ref_metadata, ref_log_events, metadata, log_events, log_path, seed
            )


class TestCaseDecoderDecompress(TestCaseDecoderBase):
    """
    Tests encoding/decoding methods against uncompressed IR stream.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = False
        self.num_test_iterations = 10
        super().setUp()


class TestCaseDecoderDecompressZstd(TestCaseDecoderBase):
    """
    Tests encoding/decoding methods against zstd compressed IR stream.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = False
        self.num_test_iterations = 10
        super().setUp()


class TestCaseDecoderDecompressDefaultQuery(TestCaseDecoderBase):
    """
    Tests encoding/decoding methods against uncompressed IR stream with the
    default empty query.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = True
        self.num_test_iterations = 1
        super().setUp()


class TestCaseDecoderDecompressZstdDefaultQuery(TestCaseDecoderBase):
    """
    Tests encoding/decoding methods against zstd compressed IR stream with the
    default empty query.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = True
        self.num_test_iterations = 1
        super().setUp()


class TestCaseDecoderTimeRangeQueryBase(TestCaseDecoderBase):
    # override
    def _generate_random_query(
        self, ref_log_events: List[LogEvent]
    ) -> Tuple[Query, List[LogEvent]]:
        self.assertGreater(len(ref_log_events), 0, "The reference log event list is empty.")
        ts_min: int = ref_log_events[0].get_timestamp()
        ts_max: int = ref_log_events[-1].get_timestamp()
        search_time_lower_bound: int = random.randint(ts_min, ts_max)
        search_time_upper_bound: int = random.randint(search_time_lower_bound, ts_max)
        query: Query = Query(
            search_time_lower_bound=search_time_lower_bound,
            search_time_upper_bound=search_time_upper_bound,
            search_time_termination_margin=0,
        )
        matched_log_events: List[LogEvent] = []
        for log_event in ref_log_events:
            if not log_event.match_query(query):
                continue
            matched_log_events.append(log_event)
        return query, matched_log_events


class TestCaseDecoderTimeRangeQuery(TestCaseDecoderTimeRangeQueryBase):
    """
    Tests encoding/decoding methods against uncompressed IR stream with the
    query that specifies a search timestamp.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseDecoderTimeRangeQueryZstd(TestCaseDecoderTimeRangeQueryBase):
    """
    Tests encoding/decoding methods against zstd compressed IR stream with the
    query that specifies a search timestamp.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseDecoderWildcardQueryBase(TestCaseDecoderBase):
    # override
    def _generate_random_query(
        self, ref_log_events: List[LogEvent]
    ) -> Tuple[Query, List[LogEvent]]:
        wildcard_queries: List[WildcardQuery] = (
            LogGenerator.generate_random_log_type_wildcard_queries(3)
        )
        query: Query = Query(wildcard_queries=wildcard_queries)
        matched_log_events: List[LogEvent] = []
        for log_event in ref_log_events:
            if not log_event.match_query(query):
                continue
            matched_log_events.append(log_event)
        return query, matched_log_events


class TestCaseDecoderWildcardQuery(TestCaseDecoderWildcardQueryBase):
    """
    Tests encoding/decoding methods against uncompressed IR stream with the
    query that specifies wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseDecoderWildcardQueryZstd(TestCaseDecoderWildcardQueryBase):
    """
    Tests encoding/decoding methods against zstd compressed IR stream with the
    query that specifies a wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseDecoderTimeRangeWildcardQueryBase(TestCaseDecoderBase):
    # override
    def _generate_random_query(
        self, ref_log_events: List[LogEvent]
    ) -> Tuple[Query, List[LogEvent]]:
        self.assertGreater(len(ref_log_events), 0, "The reference log event list is empty.")
        ts_min: int = ref_log_events[0].get_timestamp()
        ts_max: int = ref_log_events[-1].get_timestamp()
        search_time_lower_bound: int = random.randint(ts_min, ts_max)
        search_time_upper_bound: int = random.randint(search_time_lower_bound, ts_max)
        wildcard_queries: List[WildcardQuery] = (
            LogGenerator.generate_random_log_type_wildcard_queries(3)
        )
        query: Query = Query(
            search_time_lower_bound=search_time_lower_bound,
            search_time_upper_bound=search_time_upper_bound,
            wildcard_queries=wildcard_queries,
            search_time_termination_margin=0,
        )
        matched_log_events: List[LogEvent] = []
        for log_event in ref_log_events:
            if not log_event.match_query(query):
                continue
            matched_log_events.append(log_event)
        return query, matched_log_events


class TestCaseDecoderTimeRangeWildcardQuery(TestCaseDecoderTimeRangeWildcardQueryBase):
    """
    Tests encoding/decoding methods against uncompressed IR stream with the
    query that specifies both search time range and wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = False
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()


class TestCaseDecoderTimeRangeWildcardQueryZstd(TestCaseDecoderTimeRangeWildcardQueryBase):
    """
    Tests encoding/decoding methods against zstd compressed IR stream with the
    query that specifies both search time range and wildcard queries.
    """

    # override
    def setUp(self) -> None:
        self.enable_compression = True
        self.has_query = True
        self.num_test_iterations = 10
        super().setUp()



class WildcardQuery:
    """
    This class defines a wildcard query, which includes a wildcard string and a
    boolean value to indicate if the match is case-sensitive.

    A wildcard string may contain the following types of supported wildcards:

    1. '*': match 0 or more characters.
    2. '?': match any single character.

    Each wildcard can be escaped using a preceding '\\\\' (a single backslash).
    Other characters which are escaped are treated as normal characters.
    """

    def __init__(self, wildcard_query: str, case_sensitive: bool = False):
        """
        Initializes a wildcard query using the given parameters.

        :param wildcard_query: Wildcard query string.
        :param case_sensitive: Case sensitive indicator.
        """
        self._wildcard_query: str = wildcard_query
        self._case_sensitive: bool = case_sensitive

    def __str__(self) -> str:
        """
        :return: The string representation of the WildcardQuery object.
        """
        return (
            f'WildcardQuery(wildcard_query="{self._wildcard_query}",'
            f" case_sensitive={self._case_sensitive})"
        )

    def __repr__(self) -> str:
        """
        :return: Same as `__str__` method.
        """
        return self.__str__()

    @property
    def wildcard_query(self) -> str:
        return self._wildcard_query

    @property
    def case_sensitive(self) -> bool:
        return self._case_sensitive



import distutils.ccompiler
import multiprocessing.pool
import os
import platform
import sys
from setuptools import setup, Extension
from typing import List, Optional, Tuple

cpp_src_root: str = "src"
clp_src_root: str = f"{cpp_src_root}/clp/components/core/src/clp"
clp_submodule_root: str = f"{cpp_src_root}/clp/components/core/submodules"
clp_ffi_py_src_root: str = f"{cpp_src_root}/clp_ffi_py"

ir_native: Extension = Extension(
    name="clp_ffi_py.ir.native",
    language="c++",
    include_dirs=[
        cpp_src_root,
        clp_src_root,
        clp_submodule_root,
    ],
    sources=[
        f"{clp_src_root}/BufferReader.cpp",
        f"{clp_src_root}/ffi/ir_stream/decoding_methods.cpp",
        f"{clp_src_root}/ffi/ir_stream/encoding_methods.cpp",
        f"{clp_src_root}/ffi/encoding_methods.cpp",
        f"{clp_src_root}/ir/parsing.cpp",
        f"{clp_src_root}/ReaderInterface.cpp",
        f"{clp_src_root}/string_utils/string_utils.cpp",

        f"{clp_ffi_py_src_root}/ir/native/decoding_methods.cpp",
        f"{clp_ffi_py_src_root}/ir/native/encoding_methods.cpp",
        f"{clp_ffi_py_src_root}/ir/native/Metadata.cpp",
        f"{clp_ffi_py_src_root}/ir/native/PyDecoder.cpp",
        f"{clp_ffi_py_src_root}/ir/native/PyDecoderBuffer.cpp",
        f"{clp_ffi_py_src_root}/ir/native/PyFourByteEncoder.cpp",
        f"{clp_ffi_py_src_root}/ir/native/PyLogEvent.cpp",
        f"{clp_ffi_py_src_root}/ir/native/PyMetadata.cpp",
        f"{clp_ffi_py_src_root}/ir/native/PyQuery.cpp",
        f"{clp_ffi_py_src_root}/ir/native/Query.cpp",
        f"{clp_ffi_py_src_root}/modules/ir_native.cpp",
        f"{clp_ffi_py_src_root}/Py_utils.cpp",
        f"{clp_ffi_py_src_root}/utils.cpp",
    ],
    extra_compile_args=[
        "/std:c++20",
        "/O2,"
    ] if "nt" == os.name else [
        "-std=c++20",
        "-O3",
    ],
    define_macros=[("SOURCE_PATH_SIZE", str(len(os.path.abspath("./src/clp/components/core"))))],
)

def _parallel_compile(
    self: distutils.ccompiler.CCompiler,
    sources: List[str],
    output_dir: Optional[str] = None,
    macros: Optional[List[Tuple[str, Optional[str]]]] = None,
    include_dirs: Optional[List[str]] = None,
    debug: int = 0,
    extra_preargs: Optional[List[str]] = None,
    extra_postargs: Optional[List[str]] = None,
    depends: Optional[List[str]] = None,
) -> List[str]:
    """
    A replacement for distutils.ccompiler.CCompiler.compile that compiles each source file
    concurrently.
    """
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
        output_dir, macros, include_dirs, sources, depends, extra_postargs
    )
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)

    def _compile_single_file(obj: str) -> None:
        try:
            src, ext = build[obj]
        except KeyError:
            return
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

    num_cores: int = multiprocessing.cpu_count()
    with multiprocessing.pool.ThreadPool(num_cores) as pool:
        pool.map(_compile_single_file, objects)
    return objects

if "__main__" == __name__:
    try:
        if "Darwin" == platform.system():
            target_env_var_name: str = "MACOSX_DEPLOYMENT_TARGET"
            min_target_version: Tuple[int, int] = (10, 15)
            
            target_str: Optional[str] = os.environ.get(target_env_var_name)
            target: Tuple[int, ...] = ()
            if target_str is not None:
                target = tuple(int(part) for part in target_str.split("."))
            if target < min_target_version:
                target = min_target_version
            os.environ[target_env_var_name] = ".".join(str(part) for part in target)

        distutils.ccompiler.CCompiler.compile = _parallel_compile
        project_name: str = "clp_ffi_py"
        description: str = "CLP FFI Python Interface"
        extension_modules: List[Extension] = [ir_native]
        if (3, 7) > sys.version_info:
            sys.exit("The minimum Python version required is Python3.7")
        else:
            setup(
                name=project_name,
                description=description,
                ext_modules=extension_modules,
                packages=["clp_ffi_py"],
            )

    except Exception as e:
        sys.exit(f"Error: {e}")
